"""Package which helps simplifying text output generation from template. 
"""

__doc__ = '''Package which helps simplifying text output generation from template.'''


from .j2 import PrepareConfig
from .read_conditions import get_variables, get_conditions


__all__ = [
	'PrepareConfig',
	'get_conditions', 'get_variables'
	
]

__version__ = '0.0.5'




import pandas as pd
# ---------------------------------------------------------

def read_worksheet(file, wks):
	"""read the excel worksheet 
	returns -> pandas DataFrame
	"""
	try:
		df = pd.read_excel(file, sheet_name=wks).fillna("")
	except:
		raise Exception(f"CRITICAL: {file} File Read failed. Expected Worksheet named '{wks}' missing")
	return df

def read_excel(file):
	"""read the excel file, all worksheet
	returns -> dictionary with sheetname as key and pandas DataFrame as value
	"""
	dfd = pd.read_excel(file, None)
	for k,v in dfd.items():
		dfd[k] = v.fillna("")
	return dfd

def to_int(item):
	"""try to convert item to integer
	"""
	try:
		return int(item)
	except:
		return item


class DeviceDetails():
	""" Device details operations
	"""

	def __init__(self, global_file, device_file):
		self.global_regional_data_file = global_file
		self.device_filename = device_file
		self.read()

	def read(self):
		"""start executing
		"""
		self.dev_details = self.read_device()								## require
		self.wkl = self.get_wkl()											
		self.region = self.get_region()
		all_region_details = self.read_region_details()						
		self.region_details = self.merge_region_details(all_region_details)	
		self.var = self.merge_vars()										
		self.table = self.dev_details['table'].T.to_dict()					## require
		self.data = {'var': self.var, 'table': self.table}					## require

	def read_device(self):
		"""read device database
		"""
		dr = {'var':{}, 'table':None}
		dr['var']['device'] = read_worksheet(self.device_filename, 'var') 		
		dr['table'] = self.merged_table_columns()
		self.device_details = dr['var']['device']
		return dr

	def merged_table_columns(self):
		"""merge the all different interfaces/protocol etc details from individual tabs in to a single dataframe.
		"""
		dfd = read_excel(self.device_filename)
		del(dfd['var'])
		df_table = pd.DataFrame({'key':[]})
		for k, df in dfd.items():
			df_table = pd.concat([df_table, df], ignore_index=True).fillna("")
		for c in df_table:
			try:		
				df_table[c] = df_table[c].apply(to_int)
				df_table[c] = df_table[c].apply(str)
			except:
				pass
		return df_table

	def get_wkl(self):
		"""get the device location detail from device var details. 'site' field is required for the same.
		"""
		try:
			df_var = self.device_details 
			for _col in df_var.columns:
				if _col == 'var': continue
				default = df_var[df_var['var'] == 'site'][_col]
				wkl = default[default.index[0]]
				if wkl: break
		except:
			raise Exception(f"CRITICAL: Data Read failed. Expected variable 'site' missing in device 'var' Worksheet")
		if not wkl:
			raise Exception(f"CRITICAL: Data Read failed. Expected value of variable 'site' missing on device 'var' Worksheet")
		return wkl

	def get_region(self):
		"""get the region detail from device var details. 'region' field is required for the same.
		"""
		try:
			ndf = self.device_details[self.device_details['var'] == 'region']
			for _col in ndf.columns:
				if _col == 'var': continue
				return ndf[_col][ndf.index[0]].upper()
		except:
			raise Exception(f"CRITICAL: Data Read failed. Expected variable 'region' missing in device 'var' Worksheet")

	def read_region_details(self):
		"""reads global/regional data and drops other irrelavent region details from dataframe. 
		"""
		try:
			df = read_worksheet(self.global_regional_data_file, 'var') 
			for _col in df.columns:
				if _col in ('var', 'default'): continue
				if _col == self.region: continue
				df = df.drop(_col, axis=1)
			return df
		except:
			print("Global database unavailable or incorrect..")

	def _check_n_update_region(self, df):
		df.default = df[self.region] if df[self.region] != "" else df.default

	def merge_region_details(self, df):
		"""merge the region column details with defailt column, deletes region column and leave only default column along with var.
		"""
		if df is None: return pd.DataFrame()
		df.apply(self._check_n_update_region, axis=1)
		df = df.drop(self.region, axis=1)
		return df

	def merge_vars(self):
		"""merge the var details from two different dataframes ( region and device - databases )
		"""
		frames= [self.region_details, self.device_details]
		return pd.concat(frames, ignore_index=True).set_index('var').to_dict()['default']

# ---------------------------------------------------------





import ipaddress as ip
import nettoolkit as nt


def str_to_list(item):
	"""splits string and returns list, items separated by either `comma`, `enter` 
	--> list
	"""
	if isinstance(item, (str, int, float) ):
		items= []
		csv = item.strip().split(",")
		for _ in csv:
			lsv = _.strip().split("\n")
			for i in lsv:
				items.append(i)
		return items
	else:
		return item

def space_separated(items):
	"""joins provided items (iterables) by `spaces`
	--> string
	"""
	return " ".join(items)

def comma_separated(items):
	"""joins provided items (iterables) by `comma`
	--> string
	"""
	return ",".join(items)

def list_append(lst, item):
	"""append an item to list
	--> list
	"""
	return lst.append(item)

def list_extend(lst, item):
	"""Extend the list of items to list
	--> list
	"""
	return lst.extend(item)

def list_sorted(lst):
	"""provided sorted elements in list
	--> list
	"""
	return sorted(lst)

def convert_to_int(lst):
	"""convert numeric string type elements to integer type in a list.
	--> list
	"""
	return [ int(x) for x in lst]

def groups_of_nine(lst):
	"""breaks down provided list in to multiple groups with max. nine elements in each group
	--> list of lists
	"""
	# lst = np.int16(np.array(lst))
	lst = nt.LST.convert_vlans_list_to_range_of_vlans_list(lst)
	lst = [ str(_) for _ in lst ]
	return nt.LST.split(lst, 9)	

def physical_if_allowed(vlan, table):
	"""condition: checks for `filter==physical` and `vlan in vlan_members`
	--> interface value of matching row
	"""
	for key, data in table.items():
		if data['filter'].lower()=='physical' and int(vlan) in nt.LST.expand_vlan_list(str_to_list(data['vlan_members'])):
			return data['interface']
	return ""

def remove_trailing_zeros(net):
	"""removes the trailing zeros from given ipv6 address
	--> str
	"""
	while True:
		trimmers = ( "::0", ":0", "::")
		exit = True
		for t in trimmers:
			if net.endswith(t):
				net = net[:-1*len(t)]
				exit = False
		if exit: break
	return net


def ipv6_urpf_acl_network(subnet):
	pfx = ip.ip_interface(subnet)
	return str(pfx.network.network_address)

def nth_ip(net, n, withMask=False):
	"""get n-th ip address of given network.
	withMask: will return value along with mask else only subnet 
	"""
	_net = str(ip.ip_interface(net).network)
	v4 = nt.addressing(_net)
	return v4.n_thIP(n, True) if withMask else v4[n]

def mask(net):
	"""get the subnet mask for given network
	eg: 24
	"""
	_net = str(ip.ip_interface(net).network)
	v4 = nt.addressing(_net)
	return v4.mask

def netmask(net):
	"""get network mask for given network
	eg. 255.255.255.0
	"""
	return str(ip.ip_interface(net).netmask)

def invmask(net):
	"""get inverse mask for given network
	eg. 0.0.0.255
	"""
	v4 = v4addressing(net)
	return str(v4.invmask)

def addressing(net): 
	"""get the ip of given subnet
	"""
	return ip.ip_interface(net)

def int_to_str(data):
	"""get the actual physical interface value by removing training sub interfaces.
	"""
	return str(data).split(".")[0]

def v4addressing(ip, mask="32"):
	"""get the IPv4 objetct for given ip, mask (default mask=32)
	--> IPv4
	"""
	if ip.find("/") > 0: return nt.IPv4(ip)
	return nt.IPv4(ip+"/"+str(mask))

def get_summaries(lst_of_pfxs):
	"""get the summaries for provided prefixes.
	--> list
	"""
	lst_of_pfxs = nt.LST.remove_empty_members(lst_of_pfxs)
	try:
		return nt.get_summaries(*lst_of_pfxs)
	except:
		print(f"ERROR RECEIVE SUMMARIES {lst_of_pfxs}")
		return []

def iprint(x): 
	"""i print function to be use withing jinja template for debug.
	"""
	print(x)

def get_item(lst, n):
	"""get the nth item from list

	Args:
		lst (list): list containing various items
		n (int): index of item to be retrived 

	Returns:
		str: n-th item from list
	"""
	try:
		if isinstance(lst, (list, tuple)):
			return lst[n]
		else:
			return lst
	except:
		return lst





import nettoolkit as nt




def get_conditions(jinja_flie):
	"""get all conditional statements from jinja file
	--> dict
	"""
	d = {'conditions':set(), 'loops':set(), 'variables': set() }
	with open(jinja_flie, 'r') as f:
		lns = f.readlines()
	for ln in lns:
		if ln.strip().startswith("{% for "):
			d['loops'].add(ln)
		elif ln.strip().startswith("{% if") or ln.strip().startswith("{% elif"):
			d['conditions'].add(ln)
		elif ln.strip().startswith("{% set "):
			d['variables'].add(ln)
	return d

def get_variables(jinja_flie):
	"""get all jinja variables defined in jinja file
	--> set
	"""
	conds = set()
	with open(jinja_flie, 'r') as f:
		lns = f.readlines()
	for ln in lns:
		starts, ends = [], []
		for i in range(20):
			if i == 0: s,e = 0,0
			start = nt.STR.find_multi(ln, '{{', start=s, count=None, index=True, beginwith=False)
			end   = nt.STR.find_multi(ln, '}}', start=e, count=None, index=True, beginwith=False)
			if start == -1: break
			starts.append(start)
			ends.append(end)
			s = start+2
			e = end+2

		if starts == []: continue
		for s, e in zip(starts, ends):
			cond = ln[s:e+2]
			# print(cond, end='\t' )
			conds.add(cond)
		# print()
	return conds

# # -----------------------------------------------------------------------------------------




def get_model_template_version(template_file):
	"""returns the template version/model from template file if defined with variables `template_version` and `set model`.
	return dash (`-`) if none.
	"""
	template_ver, model = '-', '-'
	with open(template_file, 'r') as f:
		lns = f.readlines()
		for l in lns:
			if l.find("template_version") > 0:
				template_ver = l.split('"')[1]
			if l.find("set model") > 0:
				model = l.split('"')[1].lower()
	return model,template_ver



import jinja2
from inspect import getmembers, isfunction, isclass, isroutine

from .data_collect import DeviceDetails
from .cmn import common_fn as cmn
from . import func as func
from .func import Vrf, Vlan, Physical, Bgp, Aggregated, Loopback
from .general import *


class PrepareConfig():
	"""boiler plate code class for start configuration preparation
	"""

	# -----------------------------------------
	# IMPORT FILTERS FOR JINJA VARIABLES
	# -----------------------------------------
	filters = {}
	filters.update({'Vrf': Vrf, 'Bgp': Bgp,
		'Vlan': Vlan, 'Physical': Physical, 'Aggregated': Aggregated, 'Loopback': Loopback, 		
	})
	filters.update(dict(getmembers(cmn, isfunction)))
	filters.update(dict(getmembers(Vrf, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(Vlan, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(Physical, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(Bgp, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(Aggregated, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(Loopback, lambda x:not(isroutine(x))))['__dict__'] )
	filters.update(dict(getmembers(func, isfunction)))

	def __init__(self,
		data_file,
		jtemplate_file,
		output_folder=".",
		global_variables_file=None,	
		):
		self.data_file = data_file
		self.jtemplate_file = jtemplate_file.replace("\\", '/')
		self.output_folder = output_folder
		self.global_variables_file = global_variables_file

	def custom_class_add_to_filter(self, **kwargs):
		"""add custom classes and its methods as jinja filters. External callable.
		"""
		for filtername, _cls in kwargs.items():
			try:
				# if not self.filters.get(filtername):
				self.filters.update({filtername: _cls})
				pre = self.filters.keys()
				self.filters.update(dict(getmembers(_cls, lambda x:not(isroutine(x))))['__dict__'] )
				post = self.filters.keys()
				# print(pre==post)
				# print(getmembers(_cls, lambda x:not(isroutine(x))))
			except Exception as e:
				raise Exception(f"Class Insertion Failed for filter {filtername}\n{e}")

	def custom_module_methods_add_to_filter(self, *modules):
		"""add custom methods from module(s) as jinja filters. External callable.
		"""
		for module in modules:
			try:
				self.filters.update(dict(getmembers(module, isfunction)))
			except Exception as e:
				raise Exception(f"Module Insertion Failed {module}\n{e}")


	def start(self):
		"""kicks generation
		"""
		# ## LOAD - DATA
		DD = DeviceDetails(self.global_variables_file, self.data_file)

		# ## LOAD - JINJA TEMPLATE AND ENVIRONMENT
		templateLoader = jinja2.FileSystemLoader(searchpath='')
		templateEnv = jinja2.Environment(loader=templateLoader, 
			extensions=['jinja2.ext.loopcontrols', 'jinja2.ext.do',])
		for key, value in self.filters.items():
			templateEnv.filters[key] = value

		# ## TEMPLATE FILE		
		template = templateEnv.get_template(self.jtemplate_file)
		outputText = template.render(DD.data)#, undefined=jinja2.StrictUndefined) # Enable undefined for strict variable check

		# ## WRITE OUT
		model, template_ver = get_model_template_version(self.jtemplate_file)
		op_file = f"{self.output_folder}/{DD.data['var']['hostname']}-{model}-{template_ver}-j2Gen.cfg"
		with open(op_file, 'w') as f:
			f.write(outputText)




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="j2config",
    version="0.0.5",
    author="ALIASGAR - ALI",
    author_email="aholo2000@gmail.com",
    description="configuration generation using jinja2 - for Networking Geeks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aliasgar1978/j2config",
    # package_data={
    #     'facts_finder': [
    #         'generators/commands/*',   ## j2config not have any as such
    #         ],
    # },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=['pandas', 'jinja2', 'xlrd', 'openpyxl', 'nettoolkit'],
)




"""helper functions/classes for jinja variables ( aka: filters )
"""


# import fields
from .cmn.common_fn import *
from nettoolkit import IPv4





class Vrf():
	"""device vrf/instances

	Returns:
		Vrf: Instance of VRF

	Yields:
		Vrf: Instance of VRF
	"""	

	def __init__(self, table):
		self.table = table

	def __iter__(self):
		for key, data in self.table.items():
			if self.is_vrf(data): yield data

	@staticmethod
	def is_vrf(data):
		"""condition: `filter==vrf`

		Args:
			data (DataFrame): DataFrame containing a column `filter`

		Returns:
			bool: result of condition
		"""		
		return data['filter'].lower() == 'vrf'

	def vrf_not_none(self):
		"""condition: `vrf is not None` 

		Yields:
			data_slice: data from Row that matches condition
		"""		
		for key, data in self.table.items():
			if self.is_vrf(data) and data['vrf'] != "":
				yield data

	def sorted(self):
		"""list of available vrfs sorted by `vrf` field.

		Returns:
			list: sorted vrfs
		"""		
		self.sorted_vrf = sorted([ _['vrf']  for _ in self.vrf_not_none() ])
		return  self.sorted_vrf

	def sorted_vpnids(self):
		"""list of available vpnids sorted.

		Returns:
			list: sorted vpnids
		"""		
		self.sorted_vpnids = sorted([ int(_['vrf_vpnid'])  for _ in self.vrf_not_none() ])
		return self.sorted_vpnids

	def sorted_vrf_data(self):
		"""vrf data generator, sorted by vrf names

		Yields:
			data_slice: data for all vrf rows except vrf is none, sorted by vrf names
		"""		
		for vrf in self.sorted_vrf:
			for data in self.vrf_not_none():
				if data['vrf'] == vrf: 
					yield data
					break

	def sorted_vrf_data_by_vpnid(self):
		"""vrf data generator, sorted by vpnids

		Yields:
			data_slice: data for all vrf rows except vrf is none, sorted by vpnids
		"""		
		for vpnid in self.sorted_vpnids:
			for data in self.vrf_not_none():
				if int(data['vrf_vpnid']) == vpnid: 
					yield data
					break

	def vrf_get(self, vrf):
		"""get a particular vrf data

		Args:
			vrf (str): vrf name

		Yields:
			data_slice: data for matching vrf row 
		"""		
		for data in self.vrf_not_none():
			if data['vrf'] == vrf: 
				yield data




class Vlan():
	"""device Vlan/instances
	"""	
	def __init__(self, table):
		self.table = table

	def __iter__(self):
		for key, data in self.table.items():
			if self.is_vlan(data): yield data

	@staticmethod
	def is_vlan(data):
		"""Condition: Checks if provided data is vlan data

		Args:
			data (DataFrame): Pandas DataFrame containing `filter` column

		Returns:
			bool: result of condition
		"""		
		return data['filter'].lower() == 'vlan'


	def __vlans_range(self, start, stop):
		for data in self:
			if start <= int(data['int_number']) < stop:
				yield data

	def _sorted_vl_range(self, start, stop):
		vlans = [ int(data['int_number']) for data in self if start <= int(data['int_number']) < stop ]
		return vlans	

	def vlans_sorted_range(self, start, stop):
		"""yields data slice(s) for the vlans matching for the provided range

		Args:
			start (int): starting vlan number
			stop (int): ends vlan number

		Yields:
			data_slice: of matching vlan numbers
		"""		
		for vlan in self._sorted_vl_range(start, stop):
			for data in self:
				if start <= int(data['int_number']) < stop:
					if int(data['int_number']) == vlan:
						yield data
						break

	def vlan(self, n):
		"""returns data slice for the matching vlan number

		Args:
			n (int): vlan number

		Yields:
			data_slice: of matching vlan number
		"""		
		for data in self:
			if int(data['int_number']) == n:
				yield data
				break

	def of_instance(self, vrf):
		"""yields data slice(s) for the vrf matching with `intvrf` column

		Args:
			vrf (str): vrf name

		Yields:
			data_slice: of matching vrf with `intvrf`
		"""		
		for data in self:
			if data and data['intvrf'] == vrf: yield data


class Bgp():
	"""device Bgp/instances
	"""	

	def __init__(self, table):
		self.table = table

	def __iter__(self):
		for key, data in self.table.items():
			if self.is_bgp(data): yield data

	@staticmethod
	def is_bgp(data):
		"""Condition: Checks if provided data is bgp data

		Args:
			data (DataFrame): Pandas DataFrame containing `filter` column

		Returns:
			bool: result of condition
		"""		
		return data['filter'].lower() == 'bgp'

	def vrf_not_none(self):
		"""yields data slice(s) for the bgp information where `bgp_vrf` is not none

		Yields:
			data_slice: of matching bgp details
		"""		
		for key, data in self.table.items():
			if self.is_bgp(data) and data['bgp_vrf'] != "":
				yield data

	def bgp_peers(self, vrf):
		"""yields data slice(s) for the bgp information where `bgp_vrf` matches with provided vrf name

		Args:
			vrf (str): vrf name

		Yields:
			data_slice: of matching bgp details
		"""		
		for data in self:
			if data['bgp_vrf'] == vrf:
				yield data


class Physical():
	"""device Physical/instances
	"""	
	def __init__(self, table):
		self.table = table

	@staticmethod
	def is_physical(data):
		"""Condition: Checks if provided data is Physical Interface data

		Args:
			data (DataFrame): Pandas DataFrame containing `filter` column

		Returns:
			bool: result of condition
		"""		
		return data['filter'].lower() == 'physical'

	def __iter__(self):
		for key, data in self.table.items():
			if self.is_physical(data): yield data

	def sorted(self):
		"""provides list of sorted interface numbers

		Returns:
			list: of interface numbers sorted
		"""		
		return  sorted([ int(_['int_number'])  for _ in self ])

	def uplinks(self):
		"""yields data slice(s) for the physical interface information,
		where `int_filter` information starts with uplink.

		Yields:
			data_slice: of matching physical interfaces details
		"""		
		for data in self:
			if data['int_filter'].lower().startswith('uplink'):
				yield data		

	def sorted_interfaces(self):
		"""yields data slice(s) for the sorted physical interfaces informations

		Yields:
			data_slice: sorted physical interfaces
		"""		
		for intf in self.sorted():
			for data in self:
				if int(data['int_number']) == intf:
					yield data
					break

	def interface(self, n):
		"""yields data slice(s) for the sorted physical interfaces informations

		Args:
			n (int): interface number

		Yields:
			data_slice: of matched interface
		"""		
		for data in self:
			if int(data['int_number']) == n:
				yield data
				break

	# @staticmethod     # removed since older python not support
	def interface_type(data, intf_type):
		"""condition: is provided dataslice is of given interface type

		Args:
			data (data_slice): Pandas DataFrame slice
			intf_type (str): interface type to be verify

		Returns:
			bool: result of condition
		"""		
		return data['int_filter'].lower().startswith(intf_type)

	# @staticmethod     # removed since older python not support
	def interface_type_ends(data, x):
		"""condition: is provided dataslice ends with given argument `x`

		Args:
			data (data_slice):  Pandas DataFrame slice
			x (str): interface type ending identifier to be verify with

		Returns:
			bool: result of condition
		"""		
		return data['int_filter'].lower().endswith(x)


class Aggregated():
	"""device Aggregated/instances
	"""	
	def __init__(self, table):
		self.table = table

	def __iter__(self):
		for key, data in self.table.items():
			if self.is_aggregated(data): yield data

	@staticmethod
	def is_aggregated(data):
		"""Condition: Checks if provided data is Aggregated Interface data

		Args:
			data (DataFrame): Pandas DataFrame containing `filter` column

		Returns:
			bool: result of condition
		"""			
		return data['filter'].lower() == 'aggregated'


class Loopback():
	"""device Loopback/instances
	"""	
	def __init__(self, table):
		self.table = table

	def __iter__(self):
		for key, data in self.table.items():
			if self.is_loopback(data): yield data

	@staticmethod
	def is_loopback(data):
		"""Condition: Checks if provided data is Loopback Interface data

		Args:
			data (DataFrame): Pandas DataFrame containing `filter` column

		Returns:
			bool: result of condition
		"""				
		return data['filter'].lower() == 'loopback'



def sort(obj):
	"""exectes sorted method on provided object

	Args:
		obj (dynamic): Any object object instance declaring sorted method

	Returns:
		dynamic: sorted method output from object.
	"""	
	return obj.sorted()

